// ============================================================================
// Project:	Deductive Database
// Filename:	atomtab.cpp
// Purpose:	Atom Table (stores set of atoms, similar to Enumeration Type)
// Last Change:	17.09.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#ifdef	VER_DUMP
#include "err.h"
#endif
#include "check.h"

#include "atomtab.h"


//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

atomtab_c::atomtab_c(str_t tab_name) :
		MemPool(tab_name), LargeAtomArr(&MemPool)
{
	// Set Magic number:
	CHECK_CODE(Magic = ATOMTAB_MAGIC);

	// Check parameter:
	CHECK(tab_name != STR_NULL,
			"atomtab_c::constructor: tab_name is NULL");

	// Initialize attributes (which are not nested objects):
	MemPage  = PAGE_NULL;
	MemPtr   = ATOM_NULL;
	MemFree  = 0;
	Name     = tab_name;
	NumAtoms = 0;

	// Initialize Hash table:
	int n = ATOMTAB_HASHSIZE;
	atom_t *p = HashTab;
	while(n-- > 0)
		*p++ = ATOM_NULL;

	// Initialize array for atom by id (not really necessary):
	n = ATOMTAB_HASHSIZE;
	p = AtomArr;
	while(n-- > 0)
		*p++ = ATOM_NULL;
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

atomtab_c::~atomtab_c()
{
	// Check this object:
	CHECK_VALID("atomtab_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}

//-----------------------------------------------------------------------------
// put: Search current token in atom table, insert if not found.
//-----------------------------------------------------------------------------

int atomtab_c::put(str_t name_start, str_t name_end)
{
	// Check this object:
	CHECK_VALID("atomtab_c::put");

	// Check parameters:
	CHECK(name_start != STR_NULL, "atomtab_c::put: name_start is NULL");
	CHECK(name_end != STR_NULL, "atomtab_c::put: name_end is NULL");
	CHECK(name_start < name_end,
			"atomtab_c::put: name_start not before name_end");

	// Search in hash table. First compute hash function:
	str_t name_p = name_start;
	int hash_val = *name_p;
	if(name_p != name_end) {
		hash_val ^= (*name_p++ << 7);
		int shift = 13;
		while(name_p != name_end) {
			hash_val ^= (*name_p++ << shift);
			shift += 5;
			if(shift > static_cast<int>((sizeof(int)-1) * 8))
				shift -= (sizeof(int)-1) * 6;
		}
	}
	if(hash_val < 0)
		hash_val = -(hash_val + 1);
	CHECK(hash_val >= 0, "atomtab_c::put: Hash value is negative");
	hash_val %= ATOMTAB_HASHSIZE;

	// Now search the linked list of atoms in that entry:
	for(atom_t atom = HashTab[hash_val]; atom != ATOM_NULL;
						atom = atom->hash_next()) {
		str_t atom_name_p = atom->name();
		name_p = name_start;
		while(name_p < name_end && *atom_name_p == *name_p) {
			atom_name_p++;
			name_p++;
		}
		if(name_p == name_end && *atom_name_p == '\0')
			return atom->id();
	}

	// If we arrive here, we must construct a new atom object.

	// Get new memory page if current page has not enough space:
	if(MemFree <= 0) {
		// Allocate memory page, handle failure:
		page_t page = MemPool.alloc_page();
		if(page == PAGE_NULL)
			return -1;

		// Make new page available:
		MemPage = page;
		MemPtr  = reinterpret_cast<atom_t>(MemPage);
		MemFree = PAGE_SIZE(atom_c);
	}

	// Create Atom Object:
	CHECK(MemFree > 0, "atomtab_c::put: Page too small");
	atom_t new_atom = MemPtr++;
	MemFree--;
	new_atom->init(name_start, name_end, NumAtoms, HashTab[hash_val]);
	HashTab[hash_val] = new_atom;

	// Enter in array for atom lookup by id:
	if(NumAtoms < ATOMTAB_HASHSIZE)
		AtomArr[NumAtoms] = new_atom;
	else {
		CHECK(LargeAtomArr.length() == NumAtoms - ATOMTAB_HASHSIZE,
				"atomtab_c::put: LargeAtomArr - wrong length");
		LargeAtomArr.append(new_atom);
	}

	// Increase number of atoms (= id for next atom):
	NumAtoms++;

	// Return new atom:
	return new_atom->id();
}


//-----------------------------------------------------------------------------
// get: Return atom name for given id:
//-----------------------------------------------------------------------------

str_t atomtab_c::get(int id) {
	// Check this object:
	CHECK_VALID("atomtab_c::get");

	// Check parameters:
	CHECK(id >= 0, "atomtab_c::get: id is negative");
	CHECK(id < NumAtoms, "atomtab_c::get: id is invalid (too large)");

	// Return name of atom:
	if(id < ATOMTAB_HASHSIZE)
		return AtomArr[id]->name();
	else
		return LargeAtomArr.get(id - ATOMTAB_HASHSIZE)->name();
}


//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	atomtab_c::check() const
{
	// Check magic number:
	if(Magic != ATOMTAB_MAGIC)
		return "wrong magic number";

	// MemPtr must point into CurrPage:
	if(MemPage != PAGE_NULL &&
		MemPtr < reinterpret_cast<atom_t>(MemPage))
		return "MemPtr outside bounds (too small)";
	if(MemPage != PAGE_NULL &&
		reinterpret_cast<char*>(MemPtr) >
			reinterpret_cast<char*>(MemPage) + VER_PAGESIZE)
		return "MemPtr outside bounds (too large)";

	// MemPtr can be computed from MemFree:
	if(MemPage == PAGE_NULL && MemFree != 0)
		return "MemPage is null, but MemFree != 0";
	if(MemPage != PAGE_NULL &&
		MemPtr != reinterpret_cast<atom_t>(MemPage) +
				(PAGE_SIZE(atom_c) - MemFree))
		return "MemFree is inconsistent with MemPtr";

	// Check sub-objects:
	CHECK_SUBOBJ(MemPool, "atomtab_c::check");
	CHECK_SUBOBJ(LargeAtomArr, "atomtab_c::check");

	// Check length of flexible array for map id->atom:
	if(NumAtoms < ATOMTAB_HASHSIZE) {
		if(LargeAtomArr.length() != 0)
			return "LargeAtomArr not yet used, but length != 0";
	}
	else {
		if(LargeAtomArr.length() != NumAtoms - ATOMTAB_HASHSIZE)
			return "LargeAtomArr has wrong length";
	}

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void	atomtab_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "Atom Table";
	err_c::dump_begin(headline);

	// Compute number of entries used in the hash table:
	int entries_used = 0;
	for(int i = 0; i < ATOMTAB_HASHSIZE; i++) {
		if(HashTab[i] != ATOM_NULL)
			entries_used++;
	}
	int percent = (entries_used * 100 + (ATOMTAB_HASHSIZE/2)) /
				ATOMTAB_HASHSIZE;

	// Basic information:
	err_c::dump_str("Name:       '");
	err_c::dump_str(Name);
	err_c::dump_str("'");
	err_c::dump_nl();
	err_c::dump_str("NumAtoms:    ");
	err_c::dump_int(NumAtoms);
	err_c::dump_nl();
	err_c::dump_str("HashTabSize: ");
	err_c::dump_int(ATOMTAB_HASHSIZE);
	err_c::dump_nl();
	err_c::dump_str("EntriesUsed: ");
	err_c::dump_int(entries_used);
	err_c::dump_str(" [");
	err_c::dump_int(percent);
	err_c::dump_str("%]");
	err_c::dump_nl();
	err_c::dump_str("Memory:      ");
	int pages = MemPool.num_alloc_pages();
	err_c::dump_int(pages);
	err_c::dump_str(" Pages [");
	err_c::dump_int(PAGE_KB(pages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();
	err_c::dump_str("AtomSize:    ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(atom_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_str("AtomTabSize: ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(atomtab_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show entire hash table with chains of entries:
	err_c::dump_str("Hash Table:");
	err_c::dump_nl();
	for(int i = 0; i < ATOMTAB_HASHSIZE; i++) {
		if(HashTab[i] != ATOM_NULL) {
			err_c::dump_str("  ");
			err_c::dump_str(str_fill(i, 3));
			err_c::dump_int(i);
			err_c::dump_str(":");
			for(atom_t atom = HashTab[i]; atom != ATOM_NULL;
						atom = atom->hash_next()) {
				err_c::dump_str(" '");
				err_c::dump_str(atom->name());
				err_c::dump_str("'");
			}
			err_c::dump_nl();
		}
	}

	// Show entire list of atoms by id:
	err_c::dump_str("Atom Array (list of all atoms by id):");
	err_c::dump_nl();
	for(int i = 0; i < NumAtoms; i++) {
		err_c::dump_str("  ");
		err_c::dump_str(str_fill(i, 4));
		err_c::dump_int(i);
		err_c::dump_str(": ");
		atom_t atom;
		if(i < ATOMTAB_HASHSIZE)
			atom = AtomArr[i];
		else
			atom = LargeAtomArr.get(i - ATOMTAB_HASHSIZE);
		if(atom == ATOM_NULL)
			err_c::dump_str("*** NULL ***");
		else {
			err_c::dump_str("'");
			err_c::dump_str(atom->name());
			err_c::dump_str("'");
		}
		err_c::dump_nl();
	}

	// Done:
	err_c::dump_end();
}
#endif

